import abc
from typing import Any, Dict, List

from .errors import SnykError, SnykNotFoundError, SnykNotImplementedError
from .utils import snake_to_camel


class Manager(abc.ABC):
    def __init__(self, klass, client, instance=None):
        self.klass = klass
        self.client = client
        self.instance = instance

    @abc.abstractmethod
    def all(self):
        pass  # pragma: no cover

    def get(self, id: str):
        try:
            return next(x for x in self.all() if x.id == id)
        except StopIteration:
            raise SnykNotFoundError

    def first(self):
        try:
            return self.all()[0]
        except IndexError:
            raise SnykNotFoundError

    def _filter_by_kwargs(self, data, **kwargs: Any):
        if kwargs:
            for key, value in kwargs.items():
                data = [x for x in data if getattr(x, key) == value]
        return data

    def filter(self, **kwargs: Any):
        return self._filter_by_kwargs(self.all(), **kwargs)

    @staticmethod
    def factory(klass, client, instance=None):
        try:
            if isinstance(klass, str):
                key = klass
            else:
                key = klass.__name__
            manager = {
                "Project": ProjectManager,
                "Organization": OrganizationManager,
                "Member": MemberManager,
                "License": LicenseManager,
                "Dependency": DependencyManager,
                "Entitlement": EntitlementManager,
                "Setting": SettingManager,
                "Ignore": IgnoreManager,
                "JiraIssue": JiraIssueManager,
                "DependencyGraph": DependencyGraphManager,
                "IssueSet": IssueSetManager,
                "IssueSetAggregated": IssueSetAggregatedManager,
                "Integration": IntegrationManager,
                "IntegrationSetting": IntegrationSettingManager,
                "Tag": TagManager,
            }[key]
            return manager(klass, client, instance)
        except KeyError:
            raise SnykError


class DictManager(Manager):
    @abc.abstractmethod
    def all(self) -> Dict[str, Any]:
        pass  # pragma: no cover

    def get(self, id: str):
        try:
            return self.all()[id]
        except KeyError:
            raise SnykNotFoundError

    def filter(self, **kwargs: Any):
        raise SnykNotImplementedError

    def first(self):
        try:
            return next(iter(self.all().items()))
        except StopIteration:
            raise SnykNotFoundError


class SingletonManager(Manager):
    @abc.abstractmethod
    def all(self) -> Any:
        pass  # pragma: no cover

    def first(self):
        raise SnykNotImplementedError  # pragma: no cover

    def get(self, id: str):
        raise SnykNotImplementedError  # pragma: no cover

    def filter(self, **kwargs: Any):
        raise SnykNotImplementedError  # pragma: no cover


class OrganizationManager(Manager):
    def all(self):
        resp = self.client.get("orgs")
        orgs = []
        if "orgs" in resp.json():
            for org_data in resp.json()["orgs"]:
                orgs.append(self.klass.from_dict(org_data))
        for org in orgs:
            org.client = self.client
        return orgs


class TagManager(Manager):
    def all(self):
        return self.instance._tags

    def add(self, key, value) -> bool:
        tag = {"key": key, "value": value}
        path = "org/%s/project/%s/tags" % (
            self.instance.organization.id,
            self.instance.id,
        )
        return bool(self.client.post(path, tag))

    def delete(self, key, value) -> bool:
        tag = {"key": key, "value": value}
        path = "org/%s/project/%s/tags/remove" % (
            self.instance.organization.id,
            self.instance.id,
        )
        return bool(self.client.post(path, tag))


class ProjectManager(Manager):
    def _query(self, tags: List[Dict[str, str]] = []):
        projects = []
        if self.instance:
            path = "org/%s/projects" % self.instance.id
            if tags:
                for tag in tags:
                    if "key" not in tag or "value" not in tag or len(tag.keys()) != 2:
                        raise SnykError("Each tag must contain only a key and a value")
                data = {"filters": {"tags": {"includes": tags}}}
                resp = self.client.post(path, data)
            else:
                resp = self.client.get(path)
            if "projects" in resp.json():
                for project_data in resp.json()["projects"]:
                    project_data["organization"] = self.instance.to_dict()
                    # We move tags to _tags as a cache, to avoid the need for additional requests
                    # when working with tags. We want tags to be the manager
                    try:
                        project_data["_tags"] = project_data["tags"]
                        del project_data["tags"]
                    except KeyError:
                        pass
                    projects.append(self.klass.from_dict(project_data))
            for x in projects:
                x.organization = self.instance
        else:
            for org in self.client.organizations.all():
                projects.extend(org.projects.all())
        return projects

    def all(self):
        return self._query()

    def filter(self, tags: List[Dict[str, str]] = [], **kwargs: Any):
        if tags:
            return self._filter_by_kwargs(self._query(tags), **kwargs)
        else:
            return super().filter(**kwargs)

    def get(self, id: str):
        if self.instance:
            path = "org/%s/project/%s" % (self.instance.id, id)
            resp = self.client.get(path)
            project_data = resp.json()
            project_data["organization"] = self.instance.to_dict()
            project_klass = self.klass.from_dict(project_data)
            project_klass.organization = self.instance
            return project_klass
        else:
            return super().get(id)


class MemberManager(Manager):
    def all(self):
        path = "org/%s/members" % self.instance.id
        resp = self.client.get(path)
        members = []
        for member_data in resp.json():
            members.append(self.klass.from_dict(member_data))
        return members


class LicenseManager(Manager):
    def all(self):
        if hasattr(self.instance, "organization"):
            path = "org/%s/licenses" % self.instance.organization.id
            post_body = {"filters": {"projects": [self.instance.id]}}
        else:
            path = "org/%s/licenses" % self.instance.id
            post_body: Dict[str, Dict[str, List[str]]] = {"filters": {}}

        resp = self.client.post(path, post_body)
        license_data = resp.json()
        licenses = []
        if "results" in license_data:
            for license in license_data["results"]:
                licenses.append(self.klass.from_dict(license))
        return licenses


class DependencyManager(Manager):
    def all(self, page: int = 1):
        results_per_page = 50
        if hasattr(self.instance, "organization"):
            org_id = self.instance.organization.id
            post_body = {"filters": {"projects": [self.instance.id]}}
        else:
            org_id = self.instance.id
            post_body = {"filters": {}}

        path = "org/%s/dependencies?sortBy=dependency&order=asc&page=%s&perPage=%s" % (
            org_id,
            page,
            results_per_page,
        )

        resp = self.client.post(path, post_body)
        dependency_data = resp.json()

        total = dependency_data[
            "total"
        ]  # contains the total number of results (for pagination use)

        results = [self.klass.from_dict(item) for item in dependency_data["results"]]

        if total > (page * results_per_page):
            next_results = self.all(page + 1)
            results.extend(next_results)

        return results


class EntitlementManager(DictManager):
    def all(self) -> Dict[str, bool]:
        path = "org/%s/entitlements" % self.instance.id
        resp = self.client.get(path)
        return resp.json()


class SettingManager(DictManager):
    def all(self) -> Dict[str, Any]:
        path = "org/%s/project/%s/settings" % (
            self.instance.organization.id,
            self.instance.id,
        )
        resp = self.client.get(path)
        return resp.json()

    def update(self, **kwargs: bool) -> bool:
        path = "org/%s/project/%s/settings" % (
            self.instance.organization.id,
            self.instance.id,
        )
        post_body = {}

        settings = [
            "auto_dep_upgrade_enabled",
            "auto_dep_upgrade_ignored_dependencies",
            "auto_dep_upgrade_min_age",
            "auto_dep_upgrade_limit",
            "pull_request_fail_on_any_vulns",
            "pull_request_fail_only_for_high_severity",
            "pull_request_test_enabled",
            "pull_request_assignment",
            "pull_request_inheritance",
            "pull_request_fail_only_for_issues_with_fix",
            "auto_remediation_prs",
        ]

        for setting in settings:
            if setting in kwargs:
                post_body[snake_to_camel(setting)] = kwargs[setting]

        return bool(self.client.put(path, post_body))


class IgnoreManager(DictManager):
    def all(self) -> Dict[str, List[object]]:
        path = "org/%s/project/%s/ignores" % (
            self.instance.organization.id,
            self.instance.id,
        )
        resp = self.client.get(path)
        return resp.json()


class JiraIssueManager(DictManager):
    def all(self) -> Dict[str, List[object]]:
        path = "org/%s/project/%s/jira-issues" % (
            self.instance.organization.id,
            self.instance.id,
        )
        resp = self.client.get(path)
        return resp.json()

    def create(self, issue_id: str, fields: Any) -> Dict[str, str]:
        path = "org/%s/project/%s/issue/%s/jira-issue" % (
            self.instance.organization.id,
            self.instance.id,
            issue_id,
        )
        post_body = {"fields": fields}
        resp = self.client.post(path, post_body)
        response_data = resp.json()
        # The response we get is not following the schema as specified by the api
        # https://snyk.docs.apiary.io/#reference/projects/project-jira-issues-/create-jira-issue
        if (
            issue_id in response_data
            and len(response_data[issue_id]) > 0
            and "jiraIssue" in response_data[issue_id][0]
        ):
            return response_data[issue_id][0]["jiraIssue"]
        raise SnykError


class IntegrationManager(Manager):
    def all(self):
        path = "org/%s/integrations" % self.instance.id
        resp = self.client.get(path)
        integrations = []
        integrations_data = [{"name": x, "id": resp.json()[x]} for x in resp.json()]
        for data in integrations_data:
            integrations.append(self.klass.from_dict(data))
        for integration in integrations:
            integration.organization = self.instance
        return integrations


class IntegrationSettingManager(DictManager):
    def all(self):
        path = "org/%s/integrations/%s/settings" % (
            self.instance.organization.id,
            self.instance.id,
        )
        resp = self.client.get(path)
        return resp.json()


class DependencyGraphManager(SingletonManager):
    def all(self) -> Any:
        path = "org/%s/project/%s/dep-graph" % (
            self.instance.organization.id,
            self.instance.id,
        )
        resp = self.client.get(path)
        dependency_data = resp.json()
        if "depGraph" in dependency_data:
            return self.klass.from_dict(dependency_data["depGraph"])
        raise SnykError


class IssueSetManager(SingletonManager):
    def _convert_reserved_words(self, data):
        for key in ["vulnerabilities", "licenses"]:
            if "issues" in data and key in data["issues"]:
                for i, vuln in enumerate(data["issues"][key]):
                    if "from" in vuln:
                        data["issues"][key][i]["fromPackages"] = data["issues"][key][
                            i
                        ].pop("from")
        return data

    def all(self) -> Any:
        return self.filter()

    def filter(self, **kwargs: Any):
        path = "org/%s/project/%s/issues" % (
            self.instance.organization.id,
            self.instance.id,
        )
        filters = {
            "severities": ["high", "medium", "low"],
            "types": ["vuln", "license"],
            "ignored": False,
            "patched": False,
        }
        for filter_name in filters.keys():
            if kwargs.get(filter_name):
                filters[filter_name] = kwargs[filter_name]
        post_body = {"filters": filters}
        resp = self.client.post(path, post_body)
        return self.klass.from_dict(self._convert_reserved_words(resp.json()))


class IssueSetAggregatedManager(SingletonManager):
    def all(self) -> Any:
        return self.filter()

    def filter(self, **kwargs: Any):
        path = "org/%s/project/%s/aggregated-issues" % (
            self.instance.organization.id,
            self.instance.id,
        )
        default_filters = {
            "severities": ["high", "medium", "low"],
            "exploitMaturity": [
                "mature",
                "proof-of-concept",
                "no-known-exploit",
                "no-data",
            ],
            "types": ["vuln", "license"],
            "priority": {"score": {"min": 0, "max": 1000}},
        }

        post_body = {"filters": default_filters}

        all_filters = list(default_filters.keys()) + ["ignored", "patched"]
        for filter_name in all_filters:
            if filter_name in kwargs.keys():
                post_body["filters"][filter_name] = kwargs[filter_name]

        for optional_field in ["includeDescription", "includeIntroducedThrough"]:
            if optional_field in kwargs.keys():
                post_body[optional_field] = kwargs[optional_field]

        resp = self.client.post(path, post_body)
        return self.klass.from_dict(resp.json())



import base64
from dataclasses import InitVar, dataclass, field
from typing import Any, Dict, List, Optional, Union

import requests
from mashumaro import DataClassJSONMixin  # type: ignore

from .errors import SnykError, SnykNotImplementedError
from .managers import Manager


@dataclass
class Vulnerability(DataClassJSONMixin):
    id: str
    url: str
    title: str
    description: str
    upgradePath: List[str]
    package: str
    version: str
    severity: str
    exploitMaturity: str
    isUpgradable: bool
    isPatchable: bool
    isPinnable: bool
    identifiers: Any
    semver: Any
    fromPackages: List[str] = field(default_factory=list)
    language: Optional[str] = None
    packageManager: Optional[str] = None
    publicationTime: Optional[str] = None
    priorityScore: Optional[int] = None
    disclosureTime: Optional[str] = None
    credit: Optional[List[Any]] = field(default_factory=list)
    CVSSv3: Optional[str] = None
    cvssScore: Optional[str] = None
    ignored: Optional[List[Any]] = field(default_factory=list)
    patched: Optional[List[Any]] = field(default_factory=list)


@dataclass
class LicenseIssue(DataClassJSONMixin):
    id: str
    url: str
    title: str
    package: str
    version: str
    severity: str
    fromPackages: List[str] = field(default_factory=list)
    # Although mentioned in the schema as required, currently not returned.
    isIgnored: Optional[bool] = None
    # Although mentioned in the schema as required, currently not returned.
    isPatched: Optional[bool] = None
    language: Optional[str] = None
    priorityScore: Optional[int] = None
    packageManager: Optional[str] = None
    ignored: Optional[List[Any]] = field(default_factory=list)
    patched: Optional[List[Any]] = field(default_factory=list)


@dataclass
class Issue(DataClassJSONMixin):
    vulnerabilities: List[Vulnerability]
    licenses: List[LicenseIssue]


@dataclass
class IssueSet(DataClassJSONMixin):
    ok: bool
    packageManager: str
    dependencyCount: int
    issues: Issue


@dataclass
class IssueData(DataClassJSONMixin):
    id: str
    title: str
    severity: str
    url: str
    exploitMaturity: str
    description: Optional[str] = None
    identifiers: Optional[Any] = None
    credit: Optional[str] = None
    semver: Optional[Any] = None
    publicationTime: Optional[str] = None
    disclosureTime: Optional[str] = None
    CVSSv3: Optional[str] = None
    cvssScore: Optional[str] = None
    language: Optional[str] = None
    patches: Optional[Any] = None
    nearestFixedInVersion: Optional[str] = None


@dataclass
class AggregatedIssue(DataClassJSONMixin):
    id: str
    issueType: str
    pkgName: str
    pkgVersions: List[str]
    issueData: IssueData
    isPatched: bool
    isIgnored: bool
    introducedThrough: Optional[List[Any]] = None
    ignoreReasons: Optional[List[Any]] = None
    fixInfo: Optional[Any] = None
    # Not mentioned in schema but returned
    # https://snyk.docs.apiary.io/#reference/projects/aggregated-project-issues/list-all-aggregated-issues
    priorityScore: Optional[int] = None
    priority: Optional[Any] = None


@dataclass
class IssueSetAggregated(DataClassJSONMixin):
    issues: List[AggregatedIssue]


@dataclass
class OrganizationGroup(DataClassJSONMixin):
    name: str
    id: str


@dataclass
class Organization(DataClassJSONMixin):
    name: str
    id: str
    group: Optional[OrganizationGroup] = None
    client: InitVar[Optional[Any]] = None  # type: ignore

    @property
    def projects(self) -> Manager:
        return Manager.factory(Project, self.client, self)

    @property
    def members(self) -> Manager:
        return Manager.factory(Member, self.client, self)

    @property
    def licenses(self) -> Manager:
        return Manager.factory(License, self.client, self)

    @property
    def dependencies(self) -> Manager:
        return Manager.factory(Dependency, self.client, self)

    @property
    def entitlements(self) -> Manager:
        return Manager.factory("Entitlement", self.client, self)

    @property
    def integrations(self) -> Manager:
        return Manager.factory(Integration, self.client, self)

    """
    Imports need integrations, but exposing a high-level API that
    can find the integration from the URL of the thing you want
    to import should make things simpler. This is currently only
    supported for GitHub as an example. If this works nicely then
    extending the list of integrations, adding support for branches
    and files, and better errors, would be required.
    """

    def import_project(self, url, files: Optional[List[str]] = None) -> bool:
        try:
            components = url.split("/")
            service = components[0]
            if service == "github.com":
                owner = components[1]
                name = components[2]
                parts = name.split("@")
                branch = "master"
                if len(parts) == 2:
                    name = parts[0]
                    branch = parts[1]
            elif service == "docker.io":
                name = "/".join(components[1:])
            else:
                raise SnykNotImplementedError

        except ValueError:
            raise SnykError
        try:
            integrations = {"github.com": "github", "docker.io": "docker-hub"}
            integration_name = integrations[service]
        except KeyError:
            raise SnykError
        try:
            if service == "docker.io":
                return self.integrations.filter(name=integration_name)[0].import_image(
                    name
                )
            else:
                integration = self.integrations.filter(name=integration_name)[0]

                if files:
                    return integration.import_git(owner, name, branch, files)
                else:
                    return integration.import_git(owner, name, branch)

        except KeyError:
            raise SnykError

    # https://snyk.docs.apiary.io/#reference/users/user-organisation-notification-settings/modify-org-notification-settings
    # https://snyk.docs.apiary.io/#reference/users/user-organisation-notification-settings/get-org-notification-settings
    def notification_settings(self):
        raise SnykNotImplemented  # pragma: no cover

    # https://snyk.docs.apiary.io/#reference/organisations/the-snyk-organisation-for-a-request/invite-users
    def invite(self, email: str, admin: bool = False) -> bool:
        path = "org/%s/invite" % self.id
        payload = {"email": email, "isAdmin": admin}
        return bool(self.client.post(path, payload))

    def _test(self, path, contents=None, additional=None):
        if contents:
            # Check for a file-like object, allows us to support files
            # and strings in the same interface
            read = getattr(contents, "read", None)
            if callable(read):
                contents = contents.read()
            encoded = base64.b64encode(contents.encode()).decode()
            post_body = {
                "encoding": "base64",
                "files": {"target": {"contents": encoded}},
            }

            # Some test methods carry a second file, often a lock file
            if additional:
                read = getattr(additional, "read", None)
                if callable(read):
                    additional = additional.read()
                encoded = base64.b64encode(additional.encode()).decode()
                post_body["files"]["additional"] = {"contents": encoded}

            resp = self.client.post(path, post_body)
        else:
            resp = self.client.get(path)
        return IssueSet.from_dict(resp.json())

    def test_maven(
        self, package_group_id: str, package_artifact_id: str, version: str
    ) -> IssueSet:
        path = "test/maven/%s/%s/%s?org=%s" % (
            package_group_id,
            package_artifact_id,
            version,
            self.id,
        )
        return self._test(path)

    def test_rubygem(self, name: str, version: str) -> IssueSet:
        path = "test/rubygems/%s/%s?org=%s" % (name, version, self.id)
        return self._test(path)

    def test_python(self, name: str, version: str) -> bool:
        path = "test/pip/%s/%s?org=%s" % (name, version, self.id)
        return self._test(path)

    def test_npm(self, name: str, version: str) -> bool:
        path = "test/npm/%s/%s?org=%s" % (name, version, self.id)
        return self._test(path)

    def test_pipfile(self, contents):
        path = "test/pip?org=%s" % self.id
        return self._test(path, contents)

    def test_gemfilelock(self, contents):
        path = "test/rubygems?org=%s" % self.id
        return self._test(path, contents)

    def test_packagejson(self, contents, lock=None):
        path = "test/npm?org=%s" % self.id
        return self._test(path, contents, lock)

    def test_gradlefile(self, contents):
        path = "test/gradle?org=%s" % self.id
        return self._test(path, contents)

    def test_sbt(self, contents):
        path = "test/sbt?org=%s" % self.id
        return self._test(path, contents)

    def test_pom(self, contents):
        path = "test/maven?org=%s" % self.id
        return self._test(path, contents)

    def test_composer(self, contents, lock):
        path = "test/composer?org=%s" % self.id
        return self._test(path, contents, lock)

    def test_yarn(self, contents, lock):
        path = "test/yarn?org=%s" % self.id
        return self._test(path, contents, lock)


@dataclass
class Integration(DataClassJSONMixin):
    name: str
    id: str
    organization: Optional[Organization] = None

    @property
    def settings(self):
        if not self.organization:
            raise SnykError
        return Manager.factory("IntegrationSetting", self.organization.client, self)

    def _import(self, payload) -> bool:
        if not self.organization:
            raise SnykError
        path = "org/%s/integrations/%s/import" % (self.organization.id, self.id)
        return bool(self.organization.client.post(path, payload))

    def import_git(
        self, owner: str, name: str, branch: str = "master", files: List[str] = []
    ):
        return self._import(
            {
                "target": {"owner": owner, "name": name, "branch": branch},
                "files": [{"path": x} for x in files],
            }
        )

    def import_image(self, name: str):
        if ":" not in name:
            name = "%s:latest" % name
        return self._import({"target": {"name": name}})

    def import_gitlab(self, id: str, branch: str = "master", files: List[str] = []):
        return self._import(
            {
                "target": {"id": id, "branch": branch},
                "files": [{"path": x} for x in files],
            }
        )

    def import_bitbucket(
        self, project_key: str, name: str, repo_slug: str, files: List[str] = []
    ):
        return self._import(
            {
                "target": {
                    "projectKey": project_key,
                    "name": name,
                    "repoSlug": repo_slug,
                },
                "files": [{"path": x} for x in files],
            }
        )

    def import_heroku(self, app_id: str, slug_id: str, files: List[str] = []):
        return self._import(
            {
                "target": {"appId": app_id, "slugId": slug_id},
                "files": [{"path": x} for x in files],
            }
        )

    def import_lambda(self, function_id: str, files: List[str] = []):
        return self._import(
            {
                "target": {"functionId": function_id},
                "files": [{"path": x} for x in files],
            }
        )

    def import_cloudfoundry(self, app_id: str, files: List[str] = []):
        return self._import(
            {"target": {"appId": app_id}, "files": [{"path": x} for x in files]}
        )


@dataclass
class LicenseDependency(DataClassJSONMixin):
    id: str
    name: str
    version: str
    packageManager: str


@dataclass
class LicenseProject(DataClassJSONMixin):
    id: str
    name: str


@dataclass
class License(DataClassJSONMixin):
    id: str
    dependencies: List[LicenseDependency]
    projects: List[LicenseProject]
    severity: Optional[str] = None


@dataclass
class Member(DataClassJSONMixin):
    id: str
    username: str
    name: str
    email: str
    role: str

    # https://snyk.docs.apiary.io/#reference/organisations/manage-roles-in-organisation/update-a-member-in-the-organisation
    def update_role(self, role: str):
        raise SnykNotImplementedError  # pragma: no cover

    # https://snyk.docs.apiary.io/#reference/organisations/manage-roles-in-organisation/remove-a-member-from-the-organisation
    def delete(self):
        raise SnykNotImplementedError  # pragma: no cover


@dataclass
class IssueCounts(DataClassJSONMixin):
    low: int
    high: int
    medium: int


@dataclass
class DependencyGraphPackageInfo(DataClassJSONMixin):
    name: str
    version: Optional[str] = None


@dataclass
class DependencyGraphPackage(DataClassJSONMixin):
    id: str
    info: DependencyGraphPackageInfo


@dataclass
class Node(DataClassJSONMixin):
    nodeId: str
    pkgId: str
    deps: List[Dict[str, str]]


@dataclass
class Graph(DataClassJSONMixin):
    rootNodeId: str
    nodes: List[Node]


@dataclass
class DependencyGraph(DataClassJSONMixin):
    schemaVersion: str
    pkgManager: Dict[str, str]
    pkgs: List[DependencyGraphPackage]
    graph: Graph


@dataclass
class DependencyLicense(DataClassJSONMixin):
    id: str
    title: str
    license: str


@dataclass
class DependencyProject(DataClassJSONMixin):
    name: str
    id: str


@dataclass
class Dependency(DataClassJSONMixin):
    id: str
    name: str
    version: str
    licenses: List[DependencyLicense]
    projects: List[DependencyProject]
    latestVersion: Optional[str] = None
    latestVersionPublishedDate: Optional[str] = None
    firstPublishedDate: Optional[str] = None
    isDeprecated: Optional[bool] = None
    type: Optional[str] = None
    deprecatedVersions: Optional[List[Any]] = field(default_factory=list)
    dependenciesWithIssues: Optional[List[Any]] = field(default_factory=list)


@dataclass
class Project(DataClassJSONMixin):
    name: str
    organization: Organization
    id: str
    created: str
    origin: str
    type: str
    readOnly: bool
    testFrequency: str
    totalDependencies: int
    lastTestedDate: str
    browseUrl: str
    isMonitored: bool
    issueCountsBySeverity: IssueCounts
    imageTag: Optional[str] = None
    imageId: Optional[str] = None
    imageBaseImage: Optional[str] = None
    imagePlatform: Optional[str] = None
    imageCluster: Optional[str] = None
    hostname: Optional[str] = None
    remoteRepoUrl: Optional[str] = None
    branch: Optional[str] = None
    _tags: Optional[List[Any]] = field(default_factory=list)

    def delete(self) -> bool:
        path = "org/%s/project/%s" % (self.organization.id, self.id)
        return bool(self.organization.client.delete(path))

    @property
    def settings(self) -> Manager:
        return Manager.factory("Setting", self.organization.client, self)

    @property
    def ignores(self) -> Manager:
        return Manager.factory("Ignore", self.organization.client, self)

    @property
    def jira_issues(self) -> Manager:
        return Manager.factory("JiraIssue", self.organization.client, self)

    @property
    def dependencies(self) -> Manager:
        return Manager.factory(Dependency, self.organization.client, self)

    @property
    def licenses(self) -> Manager:
        return Manager.factory(License, self.organization.client, self)

    @property
    def dependency_graph(self) -> DependencyGraph:
        return Manager.factory(DependencyGraph, self.organization.client, self).all()

    @property
    def issueset(self) -> Manager:
        return Manager.factory(IssueSet, self.organization.client, self)

    @property
    def issueset_aggregated(self) -> Manager:
        return Manager.factory(IssueSetAggregated, self.organization.client, self)

    @property
    def vulnerabilities(self) -> List[Vulnerability]:
        return self.issueset.all().issues.vulnerabilities

    @property
    def tags(self) -> Manager:
        return Manager.factory("Tag", self.organization.client, self)

    # https://snyk.docs.apiary.io/#reference/users/user-project-notification-settings/modify-project-notification-settings
    # https://snyk.docs.apiary.io/#reference/users/user-project-notification-settings/get-project-notification-settings
    def notification_settings(self):
        raise SnykNotImplementedError  # pragma: no cover



import re

import pytest  # type: ignore

from snyk.client import SnykClient
from snyk.errors import SnykError, SnykNotFoundError, SnykNotImplementedError
from snyk.models import Integration, Member, Organization, Project


class TestModels(object):
    @pytest.fixture
    def organization(self):
        org = Organization(
            name="My Other Org", id="a04d9cbd-ae6e-44af-b573-0556b0ad4bd2"
        )
        org.client = SnykClient("token")
        return org

    @pytest.fixture
    def base_url(self):
        return "https://snyk.io/api/v1"

    @pytest.fixture
    def organization_url(self, base_url, organization):
        return "%s/org/%s" % (base_url, organization.id)


class TestOrganization(TestModels):
    @pytest.fixture
    def members(self):
        return [
            {"id": "a", "username": "b", "name": "c", "email": "d", "role": "admin"}
        ]

    @pytest.fixture
    def project(self):
        return {
            "name": "atokeneduser/goof",
            "id": "6d5813be-7e6d-4ab8-80c2-1e3e2a454545",
            "created": "2018-10-29T09:50:54.014Z",
            "origin": "cli",
            "type": "npm",
            "readOnly": "false",
            "testFrequency": "daily",
            "isMonitored": "true",
            "totalDependencies": 438,
            "issueCountsBySeverity": {"low": 8, "high": 13, "medium": 15},
            "lastTestedDate": "2019-02-05T06:21:00.000Z",
            "browseUrl": "https://app.snyk.io/org/pysnyk-test-org/project/6d5813be-7e6d-4ab8-80c2-1e3e2a454545",
            "tags": [{"key": "some-key", "value": "some-value"}],
        }

    @pytest.fixture
    def blank_test(self):
        return {
            "ok": True,
            "packageManager": "blank",
            "dependencyCount": 0,
            "issues": {"licenses": [], "vulnerabilities": []},
        }

    @pytest.fixture
    def fake_file(self):
        class FakeFile(object):
            def read(self):
                return "content"

        return FakeFile()

    def test_empty_members(self, organization, requests_mock):
        matcher = re.compile("members$")
        requests_mock.get(matcher, json=[])
        assert [] == organization.members.all()

    def test_members(self, organization, requests_mock, members):
        matcher = re.compile("members$")
        requests_mock.get(matcher, json=members)
        assert 1 == len(organization.members.all())
        assert all(type(x) is Member for x in organization.members.all())
        assert "admin" == organization.members.first().role

    def test_empty_entitlements(self, organization, requests_mock):
        matcher = re.compile("entitlements$")
        requests_mock.get(matcher, json={})
        assert {} == organization.entitlements.all()

    def test_entitlements(self, organization, requests_mock):
        matcher = re.compile("entitlements$")
        output = {"reports": True}
        requests_mock.get(matcher, json=output)
        assert output == organization.entitlements.all()

    def test_empty_integrations(self, organization, requests_mock):
        matcher = re.compile("integrations$")
        requests_mock.get(matcher, json=[])
        assert [] == organization.integrations.all()

    def test_integrations(self, organization, requests_mock):
        matcher = re.compile("integrations$")
        output = {"github": "not-a-real-id"}
        requests_mock.get(matcher, json=output)
        assert 1 == len(organization.integrations.all())
        assert all(type(x) is Integration for x in organization.integrations.all())
        assert "github" == organization.integrations.first().name

    def test_empty_licenses(self, organization, requests_mock):
        matcher = re.compile("licenses$")
        requests_mock.post(matcher, json={})
        assert [] == organization.licenses.all()

    def test_empty_dependencies(self, organization, organization_url, requests_mock):
        requests_mock.post(
            "%s/dependencies" % organization_url, json={"total": 0, "results": []}
        )
        assert [] == organization.dependencies.all()

    def test_rubygems_test(self, organization, base_url, blank_test, requests_mock):
        requests_mock.get("%s/test/rubygems/puppet/4.0.0" % base_url, json=blank_test)
        assert organization.test_rubygem("puppet", "4.0.0")

    def test_maven_test(self, organization, base_url, blank_test, requests_mock):
        requests_mock.get(
            "%s/test/maven/spring/springboot/1.0.0" % base_url, json=blank_test
        )
        assert organization.test_maven("spring", "springboot", "1.0.0")

    def test_python_test(self, organization, base_url, blank_test, requests_mock):
        requests_mock.get("%s/test/pip/django/4.0.0" % base_url, json=blank_test)
        assert organization.test_python("django", "4.0.0")

    def test_npm_test(self, organization, base_url, blank_test, requests_mock):
        requests_mock.get("%s/test/npm/snyk/1.7.100" % base_url, json=blank_test)
        assert organization.test_npm("snyk", "1.7.100")

    def test_pipfile_test_with_string(
        self, organization, base_url, blank_test, requests_mock
    ):
        requests_mock.post("%s/test/pip" % base_url, json=blank_test)
        assert organization.test_pipfile("django==4.0.0")

    def test_pipfile_test_with_file(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):
        requests_mock.post("%s/test/pip" % base_url, json=blank_test)
        assert organization.test_pipfile(fake_file)

    def test_gemfilelock_test_with_file(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):
        requests_mock.post("%s/test/rubygems" % base_url, json=blank_test)
        assert organization.test_gemfilelock(fake_file)

    def test_packagejson_test_with_file(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):

        requests_mock.post("%s/test/npm" % base_url, json=blank_test)
        assert organization.test_packagejson(fake_file)

    def test_packagejson_test_with_files(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):

        requests_mock.post("%s/test/npm" % base_url, json=blank_test)
        assert organization.test_packagejson(fake_file, fake_file)

    def test_gradlefile_test_with_file(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):

        requests_mock.post("%s/test/gradle" % base_url, json=blank_test)
        assert organization.test_gradlefile(fake_file)

    def test_sbt_test_with_file(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):

        requests_mock.post("%s/test/sbt" % base_url, json=blank_test)
        assert organization.test_sbt(fake_file)

    def test_pom_test_with_file(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):

        requests_mock.post("%s/test/maven" % base_url, json=blank_test)
        assert organization.test_pom(fake_file)

    def test_composer_with_files(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):

        requests_mock.post("%s/test/composer" % base_url, json=blank_test)
        assert organization.test_composer(fake_file, fake_file)

    def test_yarn_with_files(
        self, organization, base_url, blank_test, fake_file, requests_mock
    ):

        requests_mock.post("%s/test/yarn" % base_url, json=blank_test)
        assert organization.test_yarn(fake_file, fake_file)

    def test_missing_package_test(self, organization, base_url, requests_mock):
        requests_mock.get("%s/test/rubygems/puppet/4.0.0" % base_url, status_code=404)
        with pytest.raises(SnykError):
            organization.test_rubygem("puppet", "4.0.0")

    def test_import_git(self, organization, requests_mock):
        integration_matcher = re.compile("integrations$")
        import_matcher = re.compile("import$")
        output = {"github": "not-a-real-id"}
        requests_mock.get(integration_matcher, json=output)
        requests_mock.post(import_matcher)
        gh = organization.integrations.first()
        assert gh.import_git("org", "repo", "branch")
        payload = requests_mock.last_request.json()
        assert len(payload["files"]) == 0
        assert payload["target"]["branch"] == "branch"
        assert payload["target"]["name"] == "repo"
        assert payload["target"]["owner"] == "org"

    def test_import_project(self, organization, requests_mock):
        integration_matcher = re.compile("integrations$")
        import_matcher = re.compile("import$")
        output = {"github": "not-a-real-id"}
        requests_mock.get(integration_matcher, json=output)
        requests_mock.post(import_matcher)
        assert organization.import_project("github.com/org/repo")
        payload = requests_mock.last_request.json()
        assert len(payload["files"]) == 0
        assert payload["target"]["branch"] == "master"
        assert payload["target"]["name"] == "repo"
        assert payload["target"]["owner"] == "org"

    def test_import_project_with_files(self, organization, requests_mock):
        integration_matcher = re.compile("integrations$")
        import_matcher = re.compile("import$")
        output = {"github": "not-a-real-id"}
        requests_mock.get(integration_matcher, json=output)
        requests_mock.post(import_matcher)
        assert organization.import_project(
            "github.com/org/repo", files=["Gemfile.lock"]
        )
        payload = requests_mock.last_request.json()
        assert len(payload["files"]) == 1
        assert payload["files"][0]["path"] == "Gemfile.lock"

    def test_invite(self, organization, requests_mock):
        invite_matcher = re.compile("invite$")
        requests_mock.post(
            invite_matcher, json={"email": "example@example.com", "isAdmin": False}
        )
        assert organization.invite("example@example.com")

    def test_invite_admin(self, organization, requests_mock):
        invite_matcher = re.compile("invite$")
        requests_mock.post(
            invite_matcher, json={"email": "example@example.com", "isAdmin": True}
        )
        assert organization.invite("example@example.com", admin=True)

    def test_get_project(self, organization, project, requests_mock):
        matcher = re.compile("project/6d5813be-7e6d-4ab8-80c2-1e3e2a454545$")
        requests_mock.get(matcher, json=project)
        assert (
            "atokeneduser/goof"
            == organization.projects.get("6d5813be-7e6d-4ab8-80c2-1e3e2a454545").name
        )

    def test_get_project_organization_has_client(
        self, organization, project, requests_mock
    ):
        matcher = re.compile("project/6d5813be-7e6d-4ab8-80c2-1e3e2a454545$")
        requests_mock.get(matcher, json=project)
        assert (
            organization.projects.get(
                "6d5813be-7e6d-4ab8-80c2-1e3e2a454545"
            ).organization.client
            is not None
        )

    def test_filter_projects_by_tag_missing_value(self, organization, requests_mock):
        with pytest.raises(SnykError):
            organization.projects.filter(tags=[{"key": "some-key"}])

    def test_filter_projects_by_tag_missing_key(self, organization, requests_mock):
        with pytest.raises(SnykError):
            organization.projects.filter(tags=[{"value": "some-value"}])

    def test_filter_projects_by_tag_with_extra_key(self, organization, requests_mock):
        with pytest.raises(SnykError):
            organization.projects.filter(
                tags=[{"key": "some-key", "value": "some-value", "extra": "extra"}]
            )

    def test_filter_projects_by_tag(self, organization, requests_mock):
        tags = [{"key": "some-key", "value": "some-value"}]
        projects_matcher = re.compile("projects$")
        requests_mock.post(projects_matcher, json=[])
        organization.projects.filter(tags=tags)
        payload = requests_mock.last_request.json()
        assert payload == {"filters": {"tags": {"includes": tags}}}

    def test_filter_projects_not_by_tag(self, organization, requests_mock):
        projects_matcher = re.compile("projects$")
        requests_mock.get(projects_matcher, json=[])
        assert organization.projects.filter() == []

    def test_tags_cache(self, organization, project, requests_mock):
        projects_matcher = re.compile("projects$")
        requests_mock.get(projects_matcher, json={"projects": [project]})
        projects = organization.projects.all()
        assert projects[0]._tags == [{"key": "some-key", "value": "some-value"}]


class TestProject(TestModels):
    @pytest.fixture
    def project(self, organization):
        return Project(
            name="atokeneduser/goof",
            id="6d5813be-7e6d-4ab8-80c2-1e3e2a454545",
            created="2018-10-29T09:50:54.014Z",
            origin="cli",
            type="npm",
            readOnly="false",
            isMonitored="true",
            testFrequency="daily",
            totalDependencies=438,
            issueCountsBySeverity={"low": 8, "high": 13, "medium": 15},
            lastTestedDate="2019-02-05T06:21:00.000Z",
            browseUrl="https://app.snyk.io/org/pysnyk-test-org/project/6d5813be-7e6d-4ab8-80c2-1e3e2a454545",
            organization=organization,
        )

    @pytest.fixture
    def project_url(self, organization_url, project):
        return "%s/project/%s" % (organization_url, project.id)

    def test_delete(self, project, project_url, requests_mock):
        requests_mock.delete(project_url)
        assert project.delete()

    def test_failed_delete(self, project, project_url, requests_mock):
        requests_mock.delete(project_url, status_code=500)
        with pytest.raises(SnykError):
            project.delete()

    def test_add_tag(self, project, project_url, requests_mock):
        requests_mock.post(
            "%s/tags" % project_url, json={"key": "key", "value": "value"}
        )
        assert project.tags.add("key", "value")

    def test_delete_tag(self, project, project_url, requests_mock):
        requests_mock.post(
            "%s/tags/remove" % project_url, json={"key": "key", "value": "value"}
        )
        assert project.tags.delete("key", "value")

    def test_tags(self, project, project_url, requests_mock):
        assert [] == project.tags.all()

    def test_tags_cache(self, project, project_url, requests_mock):
        tags = [{"key": "key", "value": "value"}]
        project._tags = tags
        assert tags == project.tags.all()

    def test_empty_settings(self, project, project_url, requests_mock):
        requests_mock.get("%s/settings" % project_url, json={})
        assert {} == project.settings.all()

    def test_settings(self, project, project_url, requests_mock):
        requests_mock.get(
            "%s/settings" % project_url, json={"PullRequestTestEnabled": True}
        )
        assert 1 == len(project.settings.all())
        assert project.settings.get("PullRequestTestEnabled")

    def test_update_settings(self, project, project_url, requests_mock):
        requests_mock.put("%s/settings" % project_url)
        assert project.settings.update(pull_request_test_enabled=True)

    def test_empty_ignores(self, project, project_url, requests_mock):
        requests_mock.get("%s/ignores" % project_url, json={})
        assert {} == project.ignores.all()

    def test_ignores(self, project, project_url, requests_mock):
        requests_mock.get("%s/ignores" % project_url, json={"key": [{}]})
        assert 1 == len(project.ignores.all())
        assert [{}] == project.ignores.get("key")

    def test_missing_ignores(self, project, project_url, requests_mock):
        requests_mock.get("%s/ignores" % project_url, json={})
        with pytest.raises(SnykNotFoundError):
            project.ignores.get("not-present")

    def test_filter_not_implemented_on_dict_managers(
        self, project, project_url, requests_mock
    ):
        with pytest.raises(SnykNotImplementedError):
            project.ignores.filter(key="value")

    def test_first_fails_on_empty_dict_managers(
        self, project, project_url, requests_mock
    ):
        requests_mock.get("%s/ignores" % project_url, json={})
        with pytest.raises(SnykNotFoundError):
            project.ignores.first()

    def test_empty_jira_issues(self, project, project_url, requests_mock):
        requests_mock.get("%s/jira-issues" % project_url, json={})
        assert {} == project.jira_issues.all()

    def test_jira_issues(self, project, project_url, requests_mock):
        requests_mock.get("%s/jira-issues" % project_url, json={"key": [{}]})
        assert 1 == len(project.jira_issues.all())
        assert [{}] == project.jira_issues.get("key")

    def test_create_jira_issue(self, project, project_url, requests_mock):
        issue_id = "npm:qs:20140806-1"
        return_data = {
            "npm:qs:20140806-1": [{"jiraIssue": {"id": "10001", "key": "EX-1"}}]
        }
        adapter = requests_mock.post(
            "%s/issue/%s/jira-issue" % (project_url, issue_id), json=return_data
        )
        fields = {"summary": "something's wrong", "issuetype": {"id": "10000"}}
        out = project.jira_issues.create(issue_id, fields)
        assert adapter.last_request.json() == {"fields": fields}
        assert out["id"] == "10001"
        assert out["key"] == "EX-1"

    def test_create_jira_issue_with_error(self, project, project_url, requests_mock):
        issue_id = "npm:qs:20140806-1"
        adapter = requests_mock.post(
            "%s/issue/%s/jira-issue" % (project_url, issue_id), json={}
        )
        with pytest.raises(SnykError):
            out = project.jira_issues.create(issue_id, {})

    def test_empty_dependencies(self, project, organization_url, requests_mock):
        requests_mock.post(
            "%s/dependencies" % organization_url, json={"total": 0, "results": []}
        )
        assert [] == project.dependencies.all()

    def test_empty_issues(self, project, project_url, requests_mock):
        requests_mock.post(
            "%s/issues" % project_url,
            json={
                "ok": True,
                "packageManager": "fake",
                "dependencyCount": 0,
                "issues": {"vulnerabilities": [], "licenses": []},
            },
        )
        assert project.issueset.all().ok

    def test_empty_issues_aggregated(self, project, project_url, requests_mock):
        requests_mock.post(
            "%s/aggregated-issues" % project_url, json={"issues": []},
        )
        assert [] == project.issueset_aggregated.all().issues

    def test_empty_vulnerabilities(self, project, project_url, requests_mock):
        requests_mock.post(
            "%s/issues" % project_url,
            json={
                "ok": True,
                "packageManager": "fake",
                "dependencyCount": 0,
                "issues": {"vulnerabilities": [], "licenses": []},
            },
        )
        assert [] == project.vulnerabilities

    def test_filtering_empty_issues(self, project, project_url, requests_mock):
        requests_mock.post(
            "%s/issues" % project_url,
            json={
                "ok": True,
                "packageManager": "fake",
                "dependencyCount": 0,
                "issues": {"vulnerabilities": [], "licenses": []},
            },
        )
        assert project.issueset.filter(ignored=True).ok

    def test_filtering_empty_issues_aggregated(
        self, project, project_url, requests_mock
    ):
        requests_mock.post(
            "%s/aggregated-issues" % project_url, json={"issues": []},
        )

        assert [] == project.issueset_aggregated.filter(ignored=True).issues

    def test_empty_dependency_graph(self, project, project_url, requests_mock):
        requests_mock.get(
            "%s/dep-graph" % project_url,
            json={
                "depGraph": {
                    "schemaVersion": "fake",
                    "pkgManager": {},
                    "pkgs": [],
                    "graph": {"rootNodeId": "fake", "nodes": []},
                }
            },
        )
        assert project.dependency_graph

    def test_empty_licenses(self, project, organization_url, requests_mock):
        requests_mock.post("%s/licenses" % organization_url, json=[])
        assert [] == project.licenses.all()

    def test_empty_license_severity(
        self, organization, organization_url, requests_mock
    ):
        requests_mock.post(
            "%s/licenses" % organization_url,
            json={
                "results": [
                    {
                        "id": "MIT",
                        "dependencies": [
                            {
                                "id": "accepts@1.0.0",
                                "name": "accepts",
                                "version": "1.0.0",
                                "packageManager": "npm",
                            }
                        ],
                        "projects": [
                            {
                                "name": "atokeneduser/goof",
                                "id": "6d5813be-7e6d-4ab8-80c2-1e3e2a454545",
                            }
                        ],
                    }
                ]
            },
        )
        licenses = next(iter(organization.licenses.all()))
        assert licenses.severity is None

    def test_missing_package_version_in_dep_graph(
        self, project, project_url, requests_mock
    ):
        requests_mock.get(
            "%s/dep-graph" % project_url,
            json={
                "depGraph": {
                    "pkgManager": {"name": "fake-package-manager"},
                    "pkgs": [
                        {
                            "id": "fake-package@x.y.z",
                            "info": {"name": "fake-package-name"},
                        }
                    ],
                    "schemaVersion": "fake",
                    "graph": {"rootNodeId": "fake", "nodes": []},
                }
            },
        )
        assert next(iter(project.dependency_graph.pkgs)).info.version is None



import json

import requests


class SnykError(Exception):
    pass


class SnykHTTPError(SnykError):
    def __init__(self, resp: requests.Response):
        if hasattr(resp, "json"):
            try:
                data = resp.json()
                self.code = data.get("code")
                self.message = data.get("message")
                self.error = data.get("error")
            except json.decoder.JSONDecodeError:
                self.code = resp.status_code


class SnykNotFoundError(SnykError):
    pass


class SnykOrganizationNotFoundError(SnykError):
    pass


class SnykNotImplementedError(SnykError):
    pass



"""

Snyk API client
~~~~~~~~~~~~~~~

Snyk provides an API for various parts of the service, including accessing
project vulnerabilities, managing settings and testing individual packages
or package manifests.

    >>> import snyk
    >>> client = snyk.SnykClient("<your-snyk-api-token>")
    >>> org = client.organizations.first()
    # Return a list of Snyk Project objects
    >>> org.projects.all()
    ...
    # Return vulnerability information for dependencies from a Pipfile
    >>> handle = open("Pipfile")
    >>> org.test_pipfile(handle)
    ...
"""

from .__version__ import __description__, __license__, __title__, __url__, __version__
from .client import SnykClient



from snyk.utils import snake_to_camel


class TestUtils(object):
    def test_snake_case_to_camel(self):
        snake = "testing_this_value"
        camel = "testingThisValue"
        assert camel == snake_to_camel(snake)



import logging
from typing import Any, Optional

import requests
from retry.api import retry_call  # type: ignore

from .__version__ import __version__
from .errors import SnykHTTPError, SnykNotImplementedError
from .managers import Manager
from .models import Organization, Project

logger = logging.getLogger(__name__)


class SnykClient(object):
    API_URL = "https://snyk.io/api/v1"
    USER_AGENT = "pysnyk/%s" % __version__

    def __init__(
        self,
        token: str,
        url: Optional[str] = None,
        user_agent: Optional[str] = USER_AGENT,
        debug: bool = False,
        tries: int = 1,
        delay: int = 1,
        backoff: int = 2,
    ):
        self.api_token = token
        self.api_url = url or self.API_URL
        self.api_headers = {
            "Authorization": "token %s" % self.api_token,
            "User-Agent": user_agent,
        }
        self.api_post_headers = self.api_headers
        self.api_post_headers["Content-Type"] = "application/json"
        self.tries = tries
        self.backoff = backoff
        self.delay = delay

    def request(self, method, url: str, headers: object, json={},) -> requests.Response:
        resp = method(url, json=json, headers=headers,)
        if not resp or resp.status_code >= requests.codes.server_error:
            raise SnykHTTPError(resp)
        return resp

    def post(self, path: str, body: Any) -> requests.Response:
        url = "%s/%s" % (self.api_url, path)
        logger.debug("POST: %s" % url)
        resp = retry_call(
            self.request,
            fargs=[requests.post, url],
            fkwargs={"json": body, "headers": self.api_post_headers},
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            logger=logger,
        )
        if not resp:
            raise SnykHTTPError(resp)
        return resp

    def put(self, path: str, body: Any) -> requests.Response:
        url = "%s/%s" % (self.api_url, path)
        logger.debug("PUT: %s" % url)
        resp = retry_call(
            self.request,
            fargs=[requests.put, url],
            fkwargs={"json": body, "headers": self.api_post_headers},
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            logger=logger,
        )
        if not resp:
            raise SnykHTTPError(resp)
        return resp

    def get(self, path: str) -> requests.Response:
        url = "%s/%s" % (self.api_url, path)
        logger.debug("GET: %s" % url)
        resp = retry_call(
            self.request,
            fargs=[requests.get, url],
            fkwargs={"headers": self.api_headers},
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            logger=logger,
        )
        if not resp:
            raise SnykHTTPError(resp)
        return resp

    def delete(self, path: str) -> requests.Response:
        url = "%s/%s" % (self.api_url, path)
        logger.debug("DELETE: %s" % url)
        resp = retry_call(
            self.request,
            fargs=[requests.delete, url],
            fkwargs={"headers": self.api_headers},
            tries=self.tries,
            delay=self.delay,
            backoff=self.backoff,
            logger=logger,
        )
        if not resp:
            raise SnykHTTPError(resp)
        return resp

    @property
    def organizations(self) -> Manager:
        return Manager.factory(Organization, self)

    @property
    def projects(self) -> Manager:
        return Manager.factory(Project, self)

    # https://snyk.docs.apiary.io/#reference/general/the-api-details/get-notification-settings
    # https://snyk.docs.apiary.io/#reference/users/user-notification-settings/modify-notification-settings
    def notification_settings(self):
        raise SnykNotImplementedError  # pragma: no cover

    # https://snyk.docs.apiary.io/#reference/groups/organisations-in-groups/create-a-new-organisation-in-the-group
    # https://snyk.docs.apiary.io/#reference/0/list-members-in-a-group/list-all-members-in-a-group
    # https://snyk.docs.apiary.io/#reference/0/members-in-an-organisation-of-a-group/add-a-member-to-an-organisation-from-another-organisation-in-the-group
    def groups(self):
        raise SnykNotImplementedError  # pragma: no cover

    # https://snyk.docs.apiary.io/#reference/reporting-api/issues/get-list-of-issues
    def issues(self):
        raise SnykNotImplementedError  # pragma: no cover



import re

import pytest  # type: ignore

from snyk import SnykClient
from snyk.__version__ import __version__
from snyk.errors import SnykError, SnykNotFoundError
from snyk.models import Organization, Project


class TestSnykClient(object):
    @pytest.fixture
    def client(self):
        return SnykClient("token")

    def test_default_api_url(self, client):
        assert client.api_url == "https://snyk.io/api/v1"

    def test_overriding_api_url(self):
        url = "https://notsnyk.io/api/v1"
        client = SnykClient("token", url)
        assert client.api_url == url

    def test_token_added_to_headers(self, client):
        assert client.api_headers["Authorization"] == "token token"

    def test_user_agent_added_to_headers(self, client):
        assert client.api_headers["User-Agent"] == "pysnyk/%s" % __version__

    def test_overriding_user_agent(self):
        ua = "test"
        client = SnykClient("token", user_agent=ua)
        assert client.api_headers["User-Agent"] == ua

    def test_token_added_to_post_headers(self, client):
        assert client.api_post_headers["Authorization"] == "token token"

    def test_post_headers_use_correct_mimetype(self, client):
        assert client.api_post_headers["Content-Type"] == "application/json"

    def test_get_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.get("https://snyk.io/api/v1/sample", text="pong")
        assert client.get("sample")

    def test_put_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.put("https://snyk.io/api/v1/sample", text="pong")
        assert client.put("sample", {})

    def test_delete_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.delete("https://snyk.io/api/v1/sample")
        assert client.delete("sample")

    def test_post_sends_request_to_snyk(self, requests_mock, client):
        requests_mock.post("https://snyk.io/api/v1/sample")
        assert client.post("sample", {})
        assert requests_mock.call_count == 1

    def test_post_raises_error(self, requests_mock, client):
        requests_mock.post("https://snyk.io/api/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.post("sample", {})
        assert requests_mock.call_count == 1

    def test_put_retries_and_raises_error(self, requests_mock, client):
        requests_mock.put("https://snyk.io/api/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.put("sample", {})
        assert requests_mock.call_count == 4

    def test_delete_retries_and_raises_error(self, requests_mock, client):
        requests_mock.delete("https://snyk.io/api/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.delete("sample")
        assert requests_mock.call_count == 4

    def test_get_retries_and_raises_error(self, requests_mock, client):
        requests_mock.get("https://snyk.io/api/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.get("sample")
        assert requests_mock.call_count == 4

    def test_post_retries_and_raises_error(self, requests_mock, client):
        requests_mock.post("https://snyk.io/api/v1/sample", status_code=500, json={})
        client = SnykClient("token", tries=4, delay=0, backoff=2)
        with pytest.raises(SnykError):
            client.post("sample", {})
        assert requests_mock.call_count == 4

    def test_put_raises_error(self, requests_mock, client):
        requests_mock.put("https://snyk.io/api/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.put("sample", {})
        assert requests_mock.call_count == 1

    def test_delete_raises_error(self, requests_mock, client):
        requests_mock.delete("https://snyk.io/api/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.delete("sample")
        assert requests_mock.call_count == 1

    def test_get_raises_error(self, requests_mock, client):
        requests_mock.get("https://snyk.io/api/v1/sample", status_code=500, json={})
        with pytest.raises(SnykError):
            client.get("sample")
        assert requests_mock.call_count == 1

    def test_empty_organizations(self, requests_mock, client):
        requests_mock.get("https://snyk.io/api/v1/orgs", json={})
        assert [] == client.organizations.all()

    @pytest.fixture
    def organizations(self):
        return {
            "orgs": [
                {
                    "name": "defaultOrg",
                    "id": "689ce7f9-7943-4a71-b704-2ba575f01089",
                    "group": None,
                },
                {
                    "name": "My Other Org",
                    "id": "a04d9cbd-ae6e-44af-b573-0556b0ad4bd2",
                    "group": {
                        "name": "ACME Inc.",
                        "id": "a060a49f-636e-480f-9e14-38e773b2a97f",
                    },
                },
            ]
        }

    @pytest.fixture
    def projects(self):
        return {
            "projects": [
                {
                    "name": "atokeneduser/goof",
                    "id": "6d5813be-7e6d-4ab8-80c2-1e3e2a454545",
                    "created": "2018-10-29T09:50:54.014Z",
                    "origin": "cli",
                    "type": "npm",
                    "readOnly": "false",
                    "isMonitored": "true",
                    "testFrequency": "daily",
                    "totalDependencies": 438,
                    "issueCountsBySeverity": {"low": 8, "high": 13, "medium": 15},
                    "lastTestedDate": "2019-02-05T06:21:00.000Z",
                    "browseUrl": "https://app.snyk.io/org/pysnyk-test-org/project/6d5813be-7e6d-4ab8-80c2-1e3e2a454545",
                }
            ]
        }

    def test_loads_organizations(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        assert len(client.organizations.all()) == 2

    def test_first_organizations(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        org = client.organizations.first()
        assert "defaultOrg" == org.name

    def test_first_organizations_on_empty(self, requests_mock, client):
        requests_mock.get("https://snyk.io/api/v1/orgs", json={})
        with pytest.raises(SnykNotFoundError):
            client.organizations.first()

    def test_filter_organizations(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        assert 1 == len(client.organizations.filter(name="defaultOrg"))

    def test_filter_organizations_empty(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        assert [] == client.organizations.filter(name="not present")

    def test_loads_organization(self, requests_mock, client, organizations):
        key = organizations["orgs"][0]["id"]
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        org = client.organizations.get(key)
        assert "defaultOrg" == org.name

    def test_non_existent_organization(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        with pytest.raises(SnykNotFoundError):
            client.organizations.get("not-present")

    def test_organization_type(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        assert all(type(x) is Organization for x in client.organizations.all())

    def test_organization_attributes(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        assert client.organizations.first().name == "defaultOrg"

    def test_organization_load_group(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        assert client.organizations.all()[1].group.name == "ACME Inc."

    def test_empty_projects(self, requests_mock, client, organizations):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        matcher = re.compile("projects$")
        requests_mock.get(matcher, json={})
        assert [] == client.projects.all()

    def test_projects(self, requests_mock, client, organizations, projects):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        matcher = re.compile("projects$")
        requests_mock.get(matcher, json=projects)
        assert len(client.projects.all()) == 2
        assert all(type(x) is Project for x in client.projects.all())

    def test_project(self, requests_mock, client, organizations, projects):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        matcher = re.compile("projects$")
        requests_mock.get(matcher, json=projects)
        assert (
            "atokeneduser/goof"
            == client.projects.get("6d5813be-7e6d-4ab8-80c2-1e3e2a454545").name
        )

    def test_non_existent_project(self, requests_mock, client, organizations, projects):
        requests_mock.get("https://snyk.io/api/v1/orgs", json=organizations)
        matcher = re.compile("projects$")
        requests_mock.get(matcher, json=projects)
        with pytest.raises(SnykNotFoundError):
            client.projects.get("not-present")



import poetry_version  # type: ignore
from importlib_metadata import version  # type: ignore

__title__ = "snyk"
__description__ = "An API client for the Snyk API."
__url__ = "https://snyk.docs.apiary.io"
__license__ = "MIT"
try:
    __version__ = version("pysnyk")
except:
    __version__ = poetry_version.extract(source_file=__file__)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['snyk']

package_data = \
{'': ['*']}

install_requires = \
['importlib_metadata>=1.6.1,<2.0.0',
 'mashumaro>=1.7,<2.0',
 'poetry-version>=0.1.3,<0.2.0',
 'requests>=2.22,<3.0',
 'retry>=0.9.2,<0.10.0']

setup_kwargs = {
    'name': 'pysnyk',
    'version': '0.6.2',
    'description': 'A Python client for the Snyk API',
    'long_description': '# pysnyk\n\n[![Build Status](https://travis-ci.org/snyk-labs/pysnyk.svg?branch=master)](https://travis-ci.org/snyk-labs/pysnyk)\n\nA Python client for the [Snyk API](https://snyk.docs.apiary.io/#).\n\n## Client\n\nUsing the client requires you to provide your Snyk API token.\n\n```python\nimport snyk\nclient = snyk.SnykClient("<your-api-token>")\n```\n\nBy default the client will connect to the public Snyk service. If you are using a local installation then you can provide the API url as the second argument.\n\n```python\nimport snyk\nclient = snyk.SnykClient("<your-api-token>", "<your-instance-of-snyk>")\n```\n\nBy default the `User-Agent` string for the API requests will be `pysnyk/<version>`. If you would like to send a custom user agent you can do so as follows:\n\n```python\nimport snyk\nclient = snyk.SnykClient("<your-api-token>", user_agent="<your-instance-of-snyk>")\n```\n\nBy default the requests are not retried. If you would like to retry failed requests with `backoff` and `delay` you can do so as follows:\n\n```python\nimport snyk\nclient = snyk.SnykClient("<your-api-token>", tries=4, delay=1, backoff=2)\n```\n\n- `tries` - the maximum number of attempts. **Default:** `1` (no retries)\n- `delay` - initial delay between attempts. **Default:** `1`\n- `backoff` - multiplier applied to delay between attempts. **Default:** `2`\n\n## Organizations\n\nWith the client we can get a list of Snyk organizations you are a member of:\n\n```python\nclient.organizations.all()\n```\n\nThis returns a list of `snyk.models.Organization` objects.\n\nIf you already have the ID of the organization you\'re after you can grab it directly:\n\n```python\nclient.organizations.get("<organization-id>")\n```\n\nThis will return a single `snyk.models.Organization` object.\n\nMost of the API is scoped to organizations, so most other methods are found on the `snyk.models.Organization` objects returned by these two methods.\n\nThe `snyk.models.Organization` object has the following properties related to the API:\n\n* `entitlements` - returns the set of Snyk features available to this account\n* `dependencies`- returns a Manager for packages in use in this organization\n* `licenses` - returns a Manager for licenses currently in use by projects in this organisation\n* `members` - returns a Manager for members\n* `projects` - returns a Manager for associated projects\n* `integrations` - returns a Manager for active integrations\n\n### A note on Managers\n\nManagers provide a consistent API for accessing objects from the Snyk API. Each manager implements the following methods:\n\n* `all()` - return a list of all of the relevant objects\n* `get("<id>")` - return a single instance of the object if it exists\n* `first()` - grab the first instance of the object if one exists\n* `filter(<key>="<value>")` - return a list filtered by one or more key/value pairs\n\n### Projects\n\nOnce you have an organization you\'re likely to want to grab the related projects:\n\n```python\nclient.organizations.first().projects.all()\n```\n\nThis will return a list of `snyk.models.Project` objects.\n\nIn the case where you want to get all of the projects across all of your organizations then you can use the handy method on the client.\n\n```python\nclient.projects.all()\n```\n\nThe `snyk.models.Project` object has the following useful properties and methods:\n\n* `delete()` - deletes the project in question. Be careful as this will delete all associated data too\n* `dependencies` - returns a Manager for packages in use in this project\n* `dependency_graph` - returns a `snyk.models.DependencyGraph` object which represents the full dependency graph of package dependencies\n* `ignores` - returns a Manager for ignore rules set on the project\n* `vulnerabilities` - returns a list of `snyk.models.Vulnerability` objects with information about vulnerabilities in this project\n* `jira_issues` - returns a Manager with access to any associated Jira issues\n* `licenses` - returns a Manager for licenses currently in use by this project\n* `settings` - returns a Manager for interacting with the current project settings\n* `tags` - returns a Manager for interacting with the current project tags\n\nYou can add and delete tags using the manager:\n\n* `tags.add(key, value)` - adds a tag with the provided key/value pair to the project\n* `tags.delete(key, value)` - deletes a tag with the provided key/value pair from the project\n\nIn the case of Projects, as well as filtering by properties (as mentioned above) you can also filter by tag:\n\n```python\nclient.organizations.first().projects.filter(tags = [{"key": "some-key", "value": "some-value"}])\n```\n\nNote that the `settings` Manager can also be used to update settings like so, assuming you have a `snyk.models.Project` object in the variable `project`.\n\n```python\nproject.settings.update(pull_request_test_enabled=True)\n```\n\n### Importing new projects\n\nThe client supports a high-level `import_project` method on organizations for adding new projects to be monitored by Snyk.\n\n```python\norg = client.organizations.first()\norg.import_project("github.com/user/project@branch")\norg.import_project("docker.io/repository:tag")\n```\n\nIf you are targetting a specific manifest file or files you can pass those as an optional argument, for instance:\n\n```python\norg.import_project("github.com/user/project@branch", files=["Gemfile.lock"])\n```\n\nThis method currently only supports importing projects from GitHub and Docker Hub. For other integrations you will need to grab the lower-level `snyk.models.Integration` object from the `snyk.models.Organization.integrations` manager noted above. Other services will be added to this API soon.\n\n\n### Testing for vulnerabilities\n\nThe API also exposes methods to discover vulnerability information about individual packages. These methods are found on the Organization object.\n\n* `test_maven(<package_group_id>, <package_artifact_id>, <version>)` - returns an IssueSet containing vulnerability information for a Maven artifact\n* `test_rubygem(<name>, <version>)` - returns an IssueSet containing vulnerability information for a Ruby Gem\n* `test_python(<name>, <version>)` - returns an IssueSet containing vulnerability information for Python package from PyPi\n* `test_npm(<name>, <version>)` - returns an IssueSet containing vulnerability information for an NPM package\n\n\nHere\'s an example of checking a particular Python package.\n\n```python\n>>> org = client.organizations.first()\n>>> result = org.test_python("flask", "0.12.2")\n>>> assert result.ok\nFalse\n# You can access details of the vulnerabilities too, for example\n>>> result.issues.vulnerabilities[0].title\n\'Improper Input Validation\'\n>>> result.issues.vulnerabilities[0].identifiers\n{\'CVE\': [\'CVE-2018-1000656\'], \'CWE\': [\'CWE-20\']\n```\n\nAs well as testing individual packages you can also test all packages found in various dependency management manifests. The client currently supports the following methods:\n\n* `test_pipfile(<file-handle-or-string>)` - returns an IssueSet for all Python dependencies in a `Pipfile`\n* `test_gemfilelock(<file-handle-or-string>)` - returns an IssueSet for all Ruby dependencies in a `Gemfile`\n* `test_packagejson(<file-handle-or-string>, (<lock-file-handle-or-string>))` - returns an IssueSet for all Javascript dependencies in a `package.json` file. Optionally takes a `package.lock` file\n* `test_gradlefile(<file-handle-or-string>)` - returns an IssueSet for all dependencies in a `Gradlefile`\n* `test_sbt(<file-handle-or-string>)` - returns an IssueSet for all dependencies defined in a `.sbt` file\n* `test_pom(<file-handle-or-string>)` - returns an IssueSet for all dependencies in a Maven `pom.xml` file\n* `test_yarn(<file-handle-or-string>, <lock-file-handle-or-string>)` - returns an IssueSet for all dependencies in Yarn `package.json` and `yarn.lock` files\n* `test_composer(<file-handle-or-string>, <lock-file-handle-or-string>)` - returns an IssueSet for all dependencies in Composer `composer.json` and `composer.lock` files\n\nFor example, here we are testing a Python `Pipfile`.\n\n```python\n>>> org = client.organizations.first()\n>>> file = open("Pipfile")\n>>> org.test_pipfile(file)\n```\n\n### Inviting new users\n\nYou can invite new users to the organization via the API.\n\n```python\n>>> org = client.organizations.first()\n>>> org.invite("example@example.com")\n```\n\nYou can also invite new users as administrators:\n\n```python\n>>> org = client.organizations.first()\n>>> org.invite("example@example.com", admin=True)\n```\n\n\n### Low-level client\n\nAs well as the high-level API of the Snyk client you can use the HTTP methods directly. For these you simply need to pass the path, and optionally a data payload. The full domain, and the authentication details, are already provided by the client.\n\n```python\nclient.get("<path>")\nclient.delete("<path>")\nclient.put("<path>", <data>)\nclient.post("<path>", <data>)\n```\n\nMost of the time you shouldn\'t need to use these. They are mainly useful if new methods are added to the API which are not yet supported in the client. This can also be useful if you want to pass very specific parameters, or to parse the raw JSON output from the API.\n',
    'author': 'Gareth Rushgrove',
    'author_email': 'garethr@snyk.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/snyk-labs/pysnyk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pytest  # type: ignore

from snyk.client import SnykClient
from snyk.errors import SnykError
from snyk.managers import Manager
from snyk.models import Organization, Project


class TestManager(object):
    @pytest.fixture
    def client(self):
        return SnykClient("token")

    def test_factory(self, client):
        class NoManager(object):
            pass

        with pytest.raises(SnykError):
            Manager.factory(NoManager, client)



import re


def snake_to_camel(word):
    return lower_case_first_letter(
        "".join(x.capitalize() or "_" for x in word.split("_"))
    )


def lower_case_first_letter(word):
    return word[:1].lower() + word[1:] if word else ""
